@@ -1,4 +1,4 @@
-(* $Id: netmcore.ml 1713 2012-02-16 21:32:41Z gerd $ *)
+(* $Id: netmcore.ml 1774 2012-04-03 21:28:51Z gerd $ *)
 
 open Printf
 
@@ -20,7 +20,8 @@ type process_id =
     [ `Process of int ]
 
 type compute_resource_type =
-    [ `File | `Posix_shm | `Posix_shm_preallocated 
+    [ `File | `Posix_shm | `Posix_shm_sc | `Posix_shm_preallocated 
+    | `Posix_shm_preallocated_sc
     | `Posix_sem | `Fork_point | `Join_point
     ]
 
@@ -32,7 +33,10 @@ type inherit_request =
 type compute_resource_repr =
     [ `File of string
     | `Posix_shm of string
+    | `Posix_shm_sc of string * Netsys_sem.prefix
     | `Posix_shm_preallocated of string * Netsys_mem.memory
+    | `Posix_shm_preallocated_sc of string * Netsys_mem.memory * 
+                                       Netsys_sem.container
     | `Posix_sem of string
     | `Fork_point of (inherit_request * Netplex_encap.encap -> process_id)
     | `Join_point of (process_id -> Netplex_encap.encap option)
@@ -41,7 +45,9 @@ type compute_resource_repr =
 type trans_resource_repr =
     [ `File of string
     | `Posix_shm of string
+    | `Posix_shm_sc of string * Netsys_sem.prefix
     | `Posix_shm_preallocated
+    | `Posix_shm_preallocated_sc
     | `Posix_sem of string
     | `Fork_point
     | `Join_point
@@ -50,6 +56,7 @@ type trans_resource_repr =
 type manage_resource_repr =
     [ `File of string
     | `Posix_shm of string
+    | `Posix_shm_sc of string * Netsys_sem.prefix
     | `Posix_sem of string
     ]
 
@@ -95,7 +102,9 @@ object
   method join_res : res_id
     (* only meaningful for fork points *)
   method post_start : unit -> unit
-    (* This is run at process start time to free unneeded resources *)
+    (* This is run at process start time to free unneeded resources
+       (resources not inherited to the worker)
+     *)
   method process_body : Netplex_types.encap -> Netplex_types.encap
     (* The argument of [def_process] *)
 
@@ -152,8 +161,9 @@ module Manage_resource_lever =
 module Create_prealloc_shm_lever =
   Netplex_cenv.Make_lever
     (struct
-       type s = string * int * bool * int (* Oo.id of the container *)
-       type r = res_id * string
+       type s = string * int * bool * int (* Oo.id of the container *) * 
+                  bool
+       type r = res_id * string * string option
      end
     )
 
@@ -181,7 +191,7 @@ type levers =
       deliver : process_id * Netplex_encap.encap option -> unit;
       get_result : res_id * process_id -> Netplex_encap.encap option option option;
       manage_resource : manage_resource_repr * int -> res_id;
-      create_prealloc_shm : string * int * bool * int -> (res_id * string);
+      create_prealloc_shm : string * int * bool * int * bool -> (res_id * string * string option);
       get_resource : res_id * int -> trans_resource_repr option;
       release : res_id * int -> unit;
     }
@@ -220,17 +230,23 @@ let self_pid = ref None
 let is_worker = ref false
   (* whether this is a worker *)
 
+let enable_pmanage = ref true
+
+let pmanager = ref (Some (Netsys_pmanage.fake_pmanage()))
+
 let create_process_fwd = ref (fun _ _ _ _ _ -> assert false)
   (* defined below *)
 
-let inheritable = [ `Posix_shm_preallocated ]
+let inheritable = [ `Posix_shm_preallocated; `Posix_shm_preallocated_sc ]
 
 
 let type_of_repr =
   function
     | `File _ -> `File
     | `Posix_shm _ -> `Posix_shm
+    | `Posix_shm_sc _ -> `Posix_shm_sc
     | `Posix_shm_preallocated _ -> `Posix_shm_preallocated
+    | `Posix_shm_preallocated_sc _ -> `Posix_shm_preallocated_sc
     | `Posix_sem _ -> `Posix_sem
     | `Fork_point _ -> `Fork_point
     | `Join_point _ -> `Join_point
@@ -327,6 +343,52 @@ let master_get_result ctrl (join_point, `Process pid) =
     | Not_found -> None
 
 
+let get_pm() =
+  match !pmanager with
+    | None ->
+        Netplex_cenv.pmanage()
+    | Some pm ->
+        pm
+
+
+let pm_register res_repr =
+  let pm = get_pm() in
+  match res_repr with
+    | `File name ->
+        pm # register_file name
+    | `Posix_shm name ->
+        pm # register_posix_shm name
+    | `Posix_shm_sc(name,p) ->
+        pm # register_posix_shm name;
+        pm # register_sem_cont p
+    | `Posix_sem name ->
+        pm # register_posix_sem name
+    | `Posix_shm_preallocated(name,_) ->
+        pm # register_posix_shm name
+    | `Posix_shm_preallocated_sc(name,_,c) ->
+        pm # register_posix_shm name;
+        pm # register_sem_cont (Netsys_sem.prefix c)
+
+
+let pm_unregister res_repr =
+  let pm = get_pm() in
+  match res_repr with
+    | `File name ->
+        pm # unregister_file name
+    | `Posix_shm name ->
+        pm # unregister_posix_shm name
+    | `Posix_shm_sc(name,p) ->
+        pm # unregister_posix_shm name;
+        pm # unregister_sem_cont p
+    | `Posix_sem name ->
+        pm # unregister_posix_sem name
+    | `Posix_shm_preallocated(name,_) ->
+        pm # unregister_posix_shm name
+    | `Posix_shm_preallocated_sc(name,_,c) ->
+        pm # unregister_posix_shm name;
+        pm # unregister_sem_cont (Netsys_sem.prefix c)
+
+
 let manage_resource res_repr post_start exec =
   let res_id = !next_resid in
   incr next_resid;
@@ -337,19 +399,27 @@ let manage_resource res_repr post_start exec =
 	inherit master_resource_skel 
                   (`Resource res_id) typ (res_repr :> compute_resource_repr)
 	method destroy() =
-	  try
-	    match res_repr with
-	      | `File name ->
-		  Unix.unlink name
-	      | `Posix_shm name ->
-		  Netsys_posix.shm_unlink name
-	      | `Posix_sem name ->
-		  Netsys_posix.sem_unlink name
-	      | `Posix_shm_preallocated(name,_) ->
-		  Netsys_posix.shm_unlink name
-	  with error ->
-	    Netlog.logf `Err
-	      "Unable to destroy resource: %s" (Netexn.to_string error)
+	  ( try
+	      match res_repr with
+	        | `File name ->
+		    Unix.unlink name
+	        | `Posix_shm name ->
+		    Netsys_posix.shm_unlink name
+	        | `Posix_shm_sc(name,p) ->
+                    Netsys_sem.unlink p;
+		    Netsys_posix.shm_unlink name
+	        | `Posix_sem name ->
+		    Netsys_posix.sem_unlink name
+	        | `Posix_shm_preallocated(name,_) ->
+		    Netsys_posix.shm_unlink name
+	        | `Posix_shm_preallocated_sc(name,_,c) ->
+                    Netsys_sem.unlink (Netsys_sem.prefix c);
+		    Netsys_posix.shm_unlink name
+	    with error ->
+	      Netlog.logf `Err
+	        "Unable to destroy resource: %s" (Netexn.to_string error)
+          );
+          if !enable_pmanage then pm_unregister res_repr;
 
 	method post_start () =
 	  post_start ()
@@ -357,6 +427,7 @@ let manage_resource res_repr post_start exec =
     ) in
   Hashtbl.replace resource_table (`Resource res_id) res;
   res # used_in exec;
+  if !enable_pmanage then pm_register res_repr;
   (`Resource res_id)
 
 
@@ -364,7 +435,7 @@ let master_manage_resource ctrl (res_repr, cid) =
   manage_resource res_repr (fun _ -> ()) (`Container cid)
 
 
-let create_prealloc_shm prefix size value_area exec =
+let create_prealloc_shm prefix size value_area exec sem_flag =
   let (fd, name) =
     Netsys_posix.shm_create prefix size in
   let mem =
@@ -373,15 +444,30 @@ let create_prealloc_shm prefix size value_area exec =
   if value_area then
     Netsys_mem.value_area mem;
   dlogr (fun () -> sprintf "create_prealloc_shm %s" name);
+  let sc_opt =
+    if sem_flag then
+      Some(Netsys_sem.create_container name)
+    else
+      None in
   let post_start _ =
     Netsys_mem.memory_unmap_file mem in
+  let res =
+    match sc_opt with
+      | None -> `Posix_shm_preallocated(name, mem)
+      | Some c -> `Posix_shm_preallocated_sc(name, mem, c ) in
   let res_id =
-    manage_resource (`Posix_shm_preallocated(name, mem)) post_start exec in
-  (res_id, name)
+    manage_resource res post_start exec in
+  (res_id, name, sc_opt)
 
 
-let master_create_prealloc_shm ctrl (prefix,size,value_area,cid) =
-  create_prealloc_shm prefix size value_area (`Container cid)
+let master_create_prealloc_shm ctrl 
+                               (prefix,size,value_area,cid,sem_flag) =
+  let (res_id, name, sc_opt) =
+    create_prealloc_shm prefix size value_area (`Container cid) sem_flag
+  in
+  (res_id, name, 
+   match sc_opt with None -> None | Some c -> Some(Netsys_sem.prefix c)
+  )
 
 
 let master_get_resource ctrl (res_id, cid) =
@@ -391,10 +477,12 @@ let master_get_resource ctrl (res_id, cid) =
     let res = Hashtbl.find resource_table res_id in
     res # used_in (`Container cid);
     match res#repr with
-      | (`File _|`Posix_shm _|`Posix_sem _) as r ->
+      | (`File _|`Posix_shm _|`Posix_shm_sc _|`Posix_sem _) as r ->
 	  Some (r :> trans_resource_repr)
       | `Posix_shm_preallocated _ ->
 	  Some `Posix_shm_preallocated
+      | `Posix_shm_preallocated_sc _ ->
+	  Some `Posix_shm_preallocated_sc
       | `Fork_point _ ->
 	  Some `Fork_point
       | `Join_point _ ->
@@ -692,6 +780,9 @@ let get_resource res_id =
 	      | Some `Posix_shm_preallocated ->
 		  failwith "Netmcore.get_resource: The `Posix_shm_preallocated \
                             resource exists but is not shared with this worker"
+	      | Some `Posix_shm_preallocated_sc ->
+		  failwith "Netmcore.get_resource: The `Posix_shm_preallocated_sc \
+                            resource exists but is not shared with this worker"
 	      | Some `Fork_point ->
 		  ( object
 		      method id = res_id
@@ -763,6 +854,9 @@ let manage_file name =
 let manage_shm name =
   manage (`Posix_shm name)
 
+let manage_shm_sc name c =
+  manage (`Posix_shm_sc(name, Netsys_sem.prefix c))
+
 let manage_sem name =
   manage (`Posix_sem name)
 
@@ -775,10 +869,20 @@ let get_file res_id =
 let get_shm res_id =
   match (get_resource res_id)#repr with
     | `Posix_shm name -> name
+    | `Posix_shm_sc(name,_) -> name
     | `Posix_shm_preallocated(name,_) -> name
+    | `Posix_shm_preallocated_sc(name,_,_) -> name
     | _ ->
 	failwith "Netmcore.get_shm: the resource is not a shm object"
 
+let get_sem_container res_id =
+  match (get_resource res_id)#repr with
+    | `Posix_shm_sc(_,p) -> Netsys_sem.container p
+    | `Posix_shm_preallocated_sc(_,_,c) -> c
+    | _ ->
+	failwith "Netmcore.get_shm: the resource is not an object with \
+                  semaphore container"
+
 let get_sem res_id =
   match (get_resource res_id)#repr with
     | `Posix_sem name -> name
@@ -789,11 +893,46 @@ let create_preallocated_shm ?(value_area=false) prefix size =
   match self_exec() with
     | Some (`Container cid) ->
 	let lev = get_levers() in
-	lev.create_prealloc_shm (prefix,size,value_area,cid)
+	let (res_id,name,_) = 
+          lev.create_prealloc_shm (prefix,size,value_area,cid,false) in
+        (res_id,name)
     | Some exec ->
-	create_prealloc_shm prefix size value_area exec
+	let (res_id,name,_) =
+          create_prealloc_shm prefix size value_area exec false in
+        (res_id,name)
     | None ->
-	create_prealloc_shm prefix size value_area `Controller
+	let (res_id,name,_) =
+	  create_prealloc_shm prefix size value_area `Controller false in
+        (res_id,name)
+
+    
+let create_preallocated_shm_sc ?(value_area=false) prefix size =
+  match self_exec() with
+    | Some (`Container cid) ->
+	let lev = get_levers() in
+	let (res_id,name,p_opt) = 
+          lev.create_prealloc_shm (prefix,size,value_area,cid, true) in
+        ( match p_opt with
+            | None -> assert false
+            | Some p ->
+                let c = Netsys_sem.container p in
+                (res_id,name,c)
+        )
+    | Some exec ->
+	let (res_id,name,sc_opt) =
+          create_prealloc_shm prefix size value_area exec true in
+        ( match sc_opt with
+            | None -> assert false
+            | Some c -> (res_id,name,c)
+        )
+    | None ->
+	let (res_id,name,sc_opt) =
+	  create_prealloc_shm prefix size value_area `Controller true in
+        ( match sc_opt with
+            | None -> assert false
+            | Some c -> (res_id,name,c)
+        )
+
     
 let self_process_id() =
   match !self_pid with
@@ -815,15 +954,23 @@ let add_plugins ctrl =
 
 
 let destroy_resources () =
+  let old = !enable_pmanage in
+  enable_pmanage := false;
   Hashtbl.iter
     (fun res_id res ->
        res # destroy()
     )
-    resource_table
+    resource_table;
+  enable_pmanage := old;
+  if old then (
+    match !pmanager with
+      | None -> ()
+      | Some pm -> pm # unlink()
+  )
 
 
 let startup ~socket_directory ?pidfile ?(init_ctrl=fun _ -> ()) 
-            ~first_process
+            ?(disable_pmanage=false) ?(no_unlink=false) ~first_process
             () =
   let config_tree =
     `Section("netplex",
@@ -847,8 +994,18 @@ let startup ~socket_directory ?pidfile ?(init_ctrl=fun _ -> ())
     ~late_initializer:(fun cf ctrl ->
 			 add_plugins ctrl;
 			 init_ctrl ctrl;
+                         enable_pmanage := not disable_pmanage;
+                         ( match !pmanager with
+                             | None -> ()
+                             | Some pm ->
+                                 let real_pm = Netplex_cenv.pmanage() in
+                                 if not no_unlink then
+                                   real_pm#unlink();
+                                 real_pm # register (pm # registered);
+                                 pmanager := Some real_pm
+                         );
 			 let pid = first_process() in
-			 initial_process := Some pid
+			 initial_process := Some pid;
 		      )
     ( Netplex_mp.mp ~keep_fd_open:true ~terminate_tmo:(-1) () )
     Netplex_log.logger_factories
