@@ -1,4 +1,4 @@
-(* $Id: netmcore_camlbox.ml 1496 2010-11-27 21:49:12Z gerd $ *)
+(* $Id: netmcore_camlbox.ml 1756 2012-03-24 20:52:25Z gerd $ *)
 
 open Netcamlbox
 open Netmcore
@@ -9,8 +9,10 @@ let create_camlbox prefix n size =
     Netsys_posix.shm_create ("/" ^ prefix) 0 in
   let fd_open = ref true in
   try
+    assert(name.[0] = '/');
+    let name1 = String.sub name 1 (String.length name-1) in
     let box =
-      format_camlbox fd n size in
+      format_camlbox name1 fd n size in
     Unix.close fd;
     fd_open := false;
     let res = manage_shm name in
