@@ -1,4 +1,4 @@
-(* $Id: netplex_container.ml 1662 2011-08-29 23:05:06Z gerd $ *)
+(* $Id: netplex_container.ml 1718 2012-02-21 14:59:45Z gerd $ *)
 
 open Netplex_types
 open Netplex_ctrl_aux
@@ -17,7 +17,7 @@ let () =
 let string_of_event = 
   function
     | `event_none -> "NONE"
-    | `event_accept ->  "ACCEPT"
+    | `event_accept n ->  "ACCEPT(" ^ string_of_int n ^ ")"
     | `event_noaccept ->  "NOACCEPT"
     | `event_received_message msg ->
 	"RECEIVED_MESSAGE(" ^ msg.msg_name ^ ")"
@@ -32,17 +32,31 @@ let string_of_sys_id =
     | `Process id -> "process " ^ string_of_int id
 
 
-class std_container ?(esys = Unixqueue.create_unix_event_system()) 
+let t_poll = 0.1
+
+
+class std_container ?esys
                     ptype sockserv =
   let ssn = sockserv # name in
+  let esys =
+    match esys with
+      | Some esys -> esys
+      | None -> sockserv # processor # container_event_system() in
+  let sys_esys = Unixqueue.create_unix_event_system() in
+  let sys_mon = Uq_mt.create_monitor sys_esys in
+  let cont_thr_id = !Netsys_oothr.provider # self # id in
 object(self)
-  val sys_esys = Unixqueue.create_unix_event_system()
+  val sys_esys = sys_esys (* for subclasses *)
   val mutable rpc = None
   val mutable sys_rpc = None
   val mutable nr_conns = 0
   val mutable nr_conns_total = 0
   val mutable engines = []
   val mutable vars = Hashtbl.create 10 
+  val         vars_mutex = !Netsys_oothr.provider # create_mutex()
+  val mutable polling = None
+  val mutable polling_timer = None
+  val mutable greedy = false
 
   method container_id = (self :> container_id)
 
@@ -98,7 +112,7 @@ object(self)
       (sockserv # processor # post_start_hook)
       (self : #container :> container);
 
-    self # setup_polling();
+    self # restart_polling();
     self # protect_run();
     (* protect_run returns when all events are handled. This must include
        [rpc], so the client must now be down.
@@ -138,24 +152,68 @@ object(self)
 
   method private protect_run () =
     try 
-      Unixqueue.run esys
+      sockserv # processor # container_run esys
     with
       | error ->
 	  self # log `Crit ("run: Exception " ^ Netexn.to_string error);
 	  let b = sockserv # processor # global_exception_handler error in
 	  if b then self # protect_run()
 
+  (* The following logic means:
+     - For synchronous processors, setup_polling is always called after
+       the connection is processed, i.e. restart_polling will find that
+       polling=None in this case. Because there is no concurrency, the
+       poll loop cannot have been started before.
+     - For async processors, setup_polling is re-established after each
+       received command (e.g. an ACCEPT), and after at most t_poll seconds.
+       It is NOT immediately called when the processor is done. This
+       means that the controller sees decreased nr_conns and fully_busy params
+       after a delay of at most t_poll seconds.
+   *)
+
+  method private restart_polling() =
+    match polling with
+      | None -> 
+	  self # setup_polling()
+      | Some _ ->
+	  (* Greedy: We suppress this setup_polling, but ensure that there is a
+	     timer
+	   *)
+	  if not greedy || polling_timer = None then
+	    self # set_polling_timer();
+	  
+
+  method private set_polling_timer() =
+    let g = Unixqueue.new_group esys in
+    Unixqueue.once esys g t_poll 
+      (fun () ->
+	 self # setup_polling();
+      );
+    polling_timer <- Some g;
+    
+
+  method private reset_polling_timer() =
+    match polling_timer with
+      | None -> ()
+      | Some g ->
+	  Unixqueue.clear esys g;
+	  polling_timer <- None
+
 
   method private setup_polling() =
+    self # reset_polling_timer();
     match rpc with
       | None -> ()  (* Already shut down ! *)
       | Some r ->
+	  let fully_busy = self#conn_limit_reached in
 	  dlogr
 	    (fun () -> 
-	       sprintf "Container %d: Polling" (Oo.id self));
-	  let fully_busy = self#conn_limit_reached in
+	       sprintf "Container %d: Polling nr_conns=%d fully_busy=%B"
+		 (Oo.id self) nr_conns fully_busy);
+	  polling <- Some(nr_conns,fully_busy,engines=[]);
 	  Netplex_ctrl_clnt.Control.V1.poll'async r (nr_conns,fully_busy)
 	    (fun getreply ->
+	       polling <- None;
 	       let continue =
 		 ( try
 		     let reply = getreply() in
@@ -173,8 +231,8 @@ object(self)
                                 the server.
 			      *)
 			     false
-			 | `event_accept -> 
-			     self # enable_accepting();
+			 | `event_accept n_accept -> 
+			     self # enable_accepting n_accept;
 			     true
 			 | `event_noaccept -> 
 			     self # disable_accepting();
@@ -238,7 +296,7 @@ object(self)
 		 self # setup_polling()
 	    )
     
-  method private enable_accepting() =
+  method private enable_accepting n_accept =
     if engines = [] then (
       List.iter
 	(fun (proto, fd_array) ->
@@ -268,7 +326,8 @@ object(self)
 				     (Oo.id self)
 				     (Netsys.int64_of_file_descr fd_slave));
 			      self # disable_accepting();
-			      self # accepted fd_slave proto
+			      self # greedy_accepting
+				(n_accept - 1) [fd_slave, proto]
 			   )
 		  ~is_error:(fun err ->
 			       self # log `Crit
@@ -290,72 +349,159 @@ object(self)
     List.iter (fun e -> e # abort()) engines;
     engines <- [];
 
-  method private accepted fd_slave proto =
+  method private greedy_accepting n_accept accept_list =
+    let n_accept = ref n_accept in
+    let accept_list = ref accept_list in
+    let sockets = sockserv#sockets in
+    let x_sockets =
+      List.map
+	(fun (proto, fd_array) ->
+	   (proto, fd_array, Array.map (fun _ -> true) fd_array)
+	)
+	sockets in
     ( try
-	let proto_obj =
-	  List.find
-	    (fun proto_obj ->
-	       proto_obj#name = proto
+	let cont = ref true in
+	if !n_accept > 0 then greedy <- true;
+	while !cont && !n_accept > 0 do
+	  cont := false;
+	  List.iter
+	    (fun (proto, fd_array, fd_cont) ->
+	       Array.iteri
+		 (fun k fd ->
+		    if fd_cont.(k) then (
+		      dlogr
+			(fun () ->
+			   sprintf "Container %d: Greedy accepting on fd %Ld"
+			     (Oo.id self)
+			     (Netsys.int64_of_file_descr fd));
+		      try
+			let fd_slave, _ = Unix.accept fd in
+			cont := true;  (* try for another round *)
+			accept_list := (fd_slave, proto) :: !accept_list;
+			Unix.set_nonblock fd_slave;
+			dlogr
+			  (fun () ->
+			     sprintf "Container %d: Accepted as fd %Ld"
+			       (Oo.id self)
+			       (Netsys.int64_of_file_descr fd_slave));
+			decr n_accept;
+			if !n_accept = 0 then raise Exit;
+		      with
+			| Unix.Unix_error((Unix.EAGAIN|Unix.EWOULDBLOCK|
+					       Unix.EINTR),
+					  _,_) ->
+			    fd_cont.(k) <- false
+			| Unix.Unix_error _ as error ->
+			    self # log `Crit
+			      ("accept: Exception " ^ Netexn.to_string error);
+			    raise Exit
+		    )
+		 )
+		 fd_array
 	    )
-	    sockserv#socket_service_config#protocols in
-	if proto_obj#tcp_nodelay then
-	  Unix.setsockopt fd_slave Unix.TCP_NODELAY true
+	    x_sockets
+	done
       with
-	| Not_found -> ()
+	| Exit ->
+	    ()
     );
+    match !accept_list with
+      | [] -> ()
+      | (fd_slave_last,proto_last) :: l ->
+	  List.iter
+	    (fun (fd_slave, proto) ->
+	       self # accepted_greedy fd_slave proto
+	    )
+	    (List.rev l);
+	  (* The last connection in this round is always processed in
+	     non-greedy style, so the controller gets a notification.
+	   *)
+	  self # accepted_nongreedy fd_slave_last proto_last
+
+  method private accepted_nongreedy fd_slave proto =
+    (* We first respond with the "accepted" message to the controller.
+       This is especially important for synchronous processors, because
+       this is the last chance to notify the controller about the state
+       change in sync contexts.
+     *)
+    self # prep_socket fd_slave proto;
     match rpc with
       | None -> assert false
       | Some r ->
+	  (* Resetting polling: From this point on the container counts as
+	     busy, and we want that this state is left ASAP
+	   *)
+	  polling <- None;
 	  Rpc_client.set_batch_call r;
 	  Rpc_client.unbound_async_call
 	    r Netplex_ctrl_aux.program_Control'V1 "accepted" Xdr.XV_void
 	    (fun _ ->
-	       nr_conns <- nr_conns + 1;
-	       nr_conns_total <- nr_conns_total + 1;
-	       let regid = self # reg_conn fd_slave in
-	       let when_done_called = ref false in
-	       dlogr
-		 (fun () -> 
-		    sprintf
-		      "Container %d: processing connection on fd %Ld \
-                       (total conns: %d)" 
-		      (Oo.id self) 
-		      (Netsys.int64_of_file_descr fd_slave)
-		      nr_conns
-		 );
-	       self # workload_hook true;
-	       self # protect
-		 "process"
-		 (sockserv # processor # process
-		    ~when_done:(fun fd ->
-				  (* Note: It is up to the user to close
-                                     the descriptor. So the descriptor can
-                                     already be used for different purposes
-                                     right now!
-				   *)
-				  if not !when_done_called then (
-				    nr_conns <- nr_conns - 1;
-				    self # unreg_conn fd_slave regid;
-				    when_done_called := true;
-				    self # workload_hook false;
-				    self # setup_polling();
-				    dlogr
-				      (fun () ->
-					 sprintf "Container %d: \
+	       self # process_conn fd_slave proto
+	    )
+
+  method private accepted_greedy fd_slave proto =
+    self # prep_socket fd_slave proto;
+    self # process_conn fd_slave proto
+
+  method private process_conn fd_slave proto =
+    nr_conns <- nr_conns + 1;
+    nr_conns_total <- nr_conns_total + 1;
+    let regid = self # reg_conn fd_slave in
+    let when_done_called = ref false in
+    dlogr
+      (fun () -> 
+	 sprintf
+	   "Container %d: processing connection on fd %Ld (total conns: %d)" 
+	   (Oo.id self) 
+	   (Netsys.int64_of_file_descr fd_slave)
+	   nr_conns
+      );
+    self # workload_hook true;
+    self # protect
+      "process"
+      (sockserv # processor # process
+	 ~when_done:(fun fd ->
+		       (* Note: It is up to the user to close
+                          the descriptor. So the descriptor can
+                          already be used for different purposes
+                          right now!
+			*)
+		       if not !when_done_called then (
+			 nr_conns <- nr_conns - 1;
+			 self # unreg_conn fd_slave regid;
+			 when_done_called := true;
+			 self # workload_hook false;
+			 self # restart_polling();
+			 dlogr
+			   (fun () ->
+			      sprintf "Container %d: \
                                                Done with connection on fd %Ld \
                                                   (total conns %d)"
-					   (Oo.id self) 
-					   (Netsys.int64_of_file_descr fd_slave)
-					   nr_conns);
-				  )
-			       )
-		    (self : #container :> container)
-		    fd_slave
-		 )
-		 proto;
-	       if not !when_done_called then
-		 self # setup_polling();
-	    )
+				(Oo.id self) 
+				(Netsys.int64_of_file_descr fd_slave)
+				nr_conns);
+		       )
+		    )
+	 (self : #container :> container)
+	 fd_slave
+      )
+      proto;
+    if not !when_done_called then
+      self # restart_polling();
+
+  method private prep_socket fd_slave proto =
+    try
+      let proto_obj =
+	List.find
+	  (fun proto_obj ->
+	     proto_obj#name = proto
+	  )
+	  sockserv#socket_service_config#protocols in
+      if proto_obj#tcp_nodelay then
+	Unix.setsockopt fd_slave Unix.TCP_NODELAY true
+    with
+      | Not_found -> ()
+
 
   val mutable reg_conns = Hashtbl.create 10
   val mutable reg_conns_cnt = 0
@@ -430,17 +576,43 @@ object(self)
       | None -> failwith "#system: No RPC client available"
       | Some r -> r
 
+  method system_monitor = sys_mon
+
   method shutdown() =
     dlogr
       (fun () -> 
 	 sprintf
 	   "Container %d: shutdown" (Oo.id self));
+    (* This method can be called from a different thread. In this case,
+       we have to ensure that the shutdown code is run in the main thread
+       of the container, and then have to wait until the shutdown is
+       complete
+     *)
+    let mt_case =
+      (cont_thr_id <> !Netsys_oothr.provider # self # id) in
+    if mt_case then (
+      let mutex = !Netsys_oothr.provider # create_mutex() in
+      let cond = !Netsys_oothr.provider # create_condition() in
+      let g = Unixqueue.new_group esys in
+      Unixqueue.once esys g 0.0
+	(fun () ->
+	   self # shutdown_action();
+	   cond # signal()
+	);
+      cond # wait mutex
+    )
+    else
+      self # shutdown_action()
+    
+
+  method private shutdown_action() =
     ( try
 	Netplex_cenv.cancel_all_timers()
       with
 	  (* can happen in the admin case: *)
 	| Netplex_cenv.Not_in_container_thread -> ()
     );
+    self # reset_polling_timer();
     self # disable_accepting();
     self # disable_container_sockets();
     ( match rpc with
@@ -451,7 +623,7 @@ object(self)
 	      r
 	      self # shutdown_extra;
 	    (* We ensure that shutdown_extra is called
-               when the client is already taken down
+	       when the client is already taken down
 	     *)
     )
 
@@ -474,8 +646,13 @@ object(self)
 		  | `Notice -> log_notice
 		  | `Info -> log_info
 		  | `Debug -> log_debug in
-	      Rpc_client.set_batch_call r;
-	      Netplex_ctrl_clnt.System.V1.log r (lev,subchannel,message);
+	      Uq_mt.monitor_async 
+		sys_mon
+		(fun arg emit ->
+		   Rpc_client.set_batch_call r;
+		   Netplex_ctrl_clnt.System.V1.log'async r arg emit
+		)
+		(lev,subchannel,message);
 	    with
 	      | error ->
 		  prerr_endline("Netplex Catastrophic Error: Unable to send log message - exception " ^ Netexn.to_string error);
@@ -489,7 +666,10 @@ object(self)
     match sys_rpc with
       | None -> failwith "#lookup: No RPC client available"
       | Some r ->
-	  Netplex_ctrl_clnt.System.V1.lookup r (service,protocol)
+	  Uq_mt.monitor_async 
+	    sys_mon
+	    (Netplex_ctrl_clnt.System.V1.lookup'async r)
+	    (service,protocol)
 
   method send_message pat msg_name msg_arguments =
     dlogr
@@ -503,13 +683,20 @@ object(self)
 	    { msg_name = msg_name;
 	      msg_arguments = msg_arguments
 	    } in
-	  Netplex_ctrl_clnt.System.V1.send_message r (pat, msg)
+	  Uq_mt.monitor_async
+	    sys_mon
+	    (Netplex_ctrl_clnt.System.V1.send_message'async r)
+	    (pat, msg)
 
   method var name =
-    Hashtbl.find vars name
+    Netsys_oothr.serialize
+      vars_mutex
+      (Hashtbl.find vars) name
 
   method set_var name value =
-    Hashtbl.replace vars name value
+    Netsys_oothr.serialize
+      vars_mutex
+      (Hashtbl.replace vars name) value
 
   method call_plugin p proc_name arg =
     dlogr
@@ -527,8 +714,9 @@ object(self)
 	      | Not_found -> failwith "call_plugin: procedure not found" in
 	  let arg_str = Xdr.pack_xdr_value_as_string arg arg_ty [] in
 	  let res_str =
-	    Netplex_ctrl_clnt.System.V1.call_plugin
-	      r
+	    Uq_mt.monitor_async
+	      sys_mon
+	      (Netplex_ctrl_clnt.System.V1.call_plugin'async r)
 	      ((Int64.of_int (Oo.id p)), proc_name, arg_str) in
 	  let res = Xdr.unpack_xdr_value ~fast:true res_str res_ty [] in
 	  res
@@ -653,7 +841,10 @@ object(self)
       | Some r ->
 	  let arg_str = Marshal.to_string arg_enc [] in
 	  let res_str =
-	    Netplex_ctrl_clnt.System.V1.activate_lever r (id, arg_str) in
+	    Uq_mt.monitor_async
+	      sys_mon
+	      (Netplex_ctrl_clnt.System.V1.activate_lever'async r)
+	      (id, arg_str) in
 	  let res = Marshal.from_string res_str 0 in
 	  res
 
@@ -774,8 +965,10 @@ object(self)
     match sys_rpc with
       | None -> failwith "#register_container_socket: No RPC client available"
       | Some r ->
-	  Netplex_ctrl_clnt.System.V1.register_container_socket
-	    r (sname, pname, path)
+	  Uq_mt.monitor_async
+	    sys_mon
+	    (Netplex_ctrl_clnt.System.V1.register_container_socket'async r)
+	    (sname, pname, path)
 
   method owned_container_sockets =
     List.map
@@ -787,8 +980,10 @@ object(self)
     match sys_rpc with
       | None -> failwith "#lookup_container_sockets: No RPC client available"
       | Some r ->
-	  Netplex_ctrl_clnt.System.V1.lookup_container_sockets
-	    r (service,protocol)
+	  Uq_mt.monitor_async
+	    sys_mon
+	    (Netplex_ctrl_clnt.System.V1.lookup_container_sockets'async r)
+	    (service,protocol)
 
 
   method startup_directory =
@@ -826,7 +1021,7 @@ object(self)
     sys_rpc <- Some sys_rpc_cl;
     c_fd_clnt <- Some fd_clnt;
     c_sys_fd_clnt <- Some sys_fd_clnt;
-    self # setup_polling();
+    self # restart_polling();
 
   method private shutdown_extra() =
     (* In the admin case, fd_clnt and sys_fd_clnt are never closed.
