@@ -1,4 +1,4 @@
-(* $Id: netplex_types.mli 1662 2011-08-29 23:05:06Z gerd $ *)
+(* $Id: netplex_types.mli 1718 2012-02-21 14:59:45Z gerd $ *)
 
 (** Types for [Netplex] *)
 
@@ -100,17 +100,19 @@ type container_state =
    *)
 
 type capacity =
-    [ `Normal_quality of int
-    | `Low_quality of int
+    [ `Normal_quality of int * bool
+    | `Low_quality of int * bool
     | `Unavailable
     ]
   (** How many connections a container can accept in addition to the
     * existing connections:
-    * - [`Normal_quality n]: It can accept n connections with normal
+    * - [`Normal_quality(n,greedy)]: It can accept n connections with normal
     *   service quality, [n > 0]
-    * - [`Low_quality n]: It can accept n connections with low
+    * - [`Low_quality(n,greedy)]: It can accept n connections with low
     *   service quality (e.g. because it is already quite loaded), [n > 0]
     * - [`Unavailable]: No capacity free
+    *
+    * The [greedy] flag sets whether greedy accepts are allowed.
    *)
 
 type extended_address =
@@ -519,6 +521,18 @@ object
       * the container continues running.
      *)
 
+  method container_event_system : unit -> Unixqueue.event_system
+    (** This method is called to get the event systems for containers.
+	This is normally a {!Unixqueue.standard_event_system}, but
+	users can override it.
+     *)
+
+  method container_run : Unixqueue.event_system -> unit
+    (** [container_run esys]: By default, it just runs [esys#run()].
+	This method is called to run the event system of the containers.
+	Users can override it.
+     *)
+
 end
 
 (** The processor is the object that is notified when a new TCP connection
@@ -553,6 +567,9 @@ end
 
 (** Containers encapsulate the control flow of the service components.
     A container is run in a separate thread or process.
+
+    {b Thread safety:} All methods except [start] can be called from
+    any thread, and provide full thread safety.
  *)
 and container =
 object
@@ -593,8 +610,14 @@ object
     (** An RPC client that can be used to send messages to the controller.
       * Only available while [start] is running. It is bound to 
       * [System.V1].
+      *
+      * In multi-threaded programs access to [system] must be governed
+      * by [system_monitor]. See {!Uq_mt} for details what this means.
      *)
 
+  method system_monitor : Uq_mt.monitor
+    (** The thread monitor protecting the [system] RPC client *)
+
   method lookup : string -> string -> string option
     (** [lookup service_name protocol_name] tries to find a Unix domain
       * socket for the service and returns it.
