@@ -1,4 +1,4 @@
-(* $Id: nethttp.mlp 1588 2011-04-28 13:59:54Z gerd $ 
+(* $Id: nethttp.mlp 1787 2012-06-27 14:28:10Z gerd $ 
  * ----------------------------------------------------------------------
  * Nethttp: Basic definitions for the HTTP protocol
  *)
@@ -992,41 +992,27 @@ module Header =
       in mh#update_field "Accept" s
       
     let best_media_type mh supp =
-      let supp' = (* All of [supp] not mentioned in the [Accept] field *)
-        let toks = try get_accept mh with | Not_found -> []
+      let match_mime a b =
+        let (main_type, sub_type) = Mimestring.split_mime_type b
         in
-          List.filter
-            (fun supp_type ->
-               not (List.exists (fun (t, _, _) -> t = supp_type) toks))
-            supp in
-      let rec find_best toks =
-        match toks with
-        | (tok, params, qparams) :: toks' ->
-            if List.mem tok supp
-            then (tok, params)
-            else
-              (let (main_type, sub_type) = Mimestring.split_mime_type tok
-               in
-                 if sub_type = "*"
-                 then
-                   (try
-                      ((List.find
-                          (fun supp_type ->
-                             (main_type = "*") ||
-                               ((sub_type = "*") &&
-                                  (main_type =
-                                     (fst
-                                        (Mimestring.split_mime_type supp_type)))))
-                          supp'),
-                       params)
-                    with | Not_found -> find_best toks')
-                 else find_best toks')
-        | [] -> (* Nothing acceptable: *) ("", [])
+          (sub_type = "*") && (*Ignore non-wildcard types*)
+            ((main_type = "*") ||
+               (main_type = (fst (Mimestring.split_mime_type a)))) in
+      let filter p l =
+        List.fold_right
+          (fun (((tok, _, _) as e)) l -> if p tok then e :: l else l) l [] in
+      let accept = try get_accept mh with | Not_found -> [ ("*/*", [], []) ]
       in
-        try
-          let mt_list = sort_by_q' (get_accept mh)
-          in (* or Not_found *) find_best mt_list
-        with | Not_found -> ("*/*", [])
+        match sort_by_q'
+                (List.flatten
+                   (List.map
+                      (fun t ->
+                         (filter (( = ) t) accept) @
+                           (filter (match_mime t) accept))
+                      supp))
+        with
+        | (tok, params, qparams) :: _ -> (tok, params)
+        | [] -> ("", [])
       
     let get_accept_charset mh =
       parse_parameterized_token_list mh "Nethttp.get_accept_charset"
