@@ -1,7 +1,8 @@
 #define HAVE_PTHREAD
 #define FANCY_PAGE_TABLES
 #define HAVE_POSIX_SHM
-#define HAVE_POSIX_SEM
+#define HAVE_POSIX_SEM_ANON
+#define HAVE_POSIX_SEM_NAMED
 #define HAVE_POSIX_SPAWN
 #define HAVE_POSIX_FADVISE
 #define HAVE_POSIX_FALLOCATE
