@@ -1,4 +1,4 @@
-/* $Id: netsys_c.c 1691 2012-02-05 18:29:36Z gerd $
+/* $Id: netsys_c.c 1790 2012-07-17 14:14:14Z gerd $
  * ----------------------------------------------------------------------
  */
 
@@ -447,7 +447,7 @@ static int at_flags_table[] = {
 
 static int open_flag_table[] = {
   O_RDONLY, O_WRONLY, O_RDWR, O_NONBLOCK, O_APPEND, O_CREAT, O_TRUNC, O_EXCL, 
-  O_NOCTTY, O_DSYNC, O_SYNC, O_RSYNC
+  O_NOCTTY, O_DSYNC, O_SYNC, O_RSYNC, 0 /* O_SHARE_DELETE */
 };
 #endif
 
