@@ -1,4 +1,4 @@
-/* $Id: netsys_c_mem.c 1716 2012-02-20 16:59:27Z gerd $ */
+/* $Id: netsys_c_mem.c 1789 2012-06-27 14:31:32Z gerd $ */
 
 #include "netsys_c.h"
 #include "netsys_c_htab.h"
@@ -80,13 +80,14 @@ CAMLprim value netsys_grab(value addrv, value lenv)
 }
 
 
-CAMLprim value netsys_alloc_memory_pages(value addrv, value pv)
+CAMLprim value netsys_alloc_memory_pages(value addrv, value pv, value execv)
 {
 #if defined(HAVE_MMAP) && defined(HAVE_SYSCONF) && defined(MAP_ANON)
     void *start;
     size_t length;
     long pgsize;
     void *data;
+    int flags;
     value r;
 
     start = (void *) Nativeint_val(addrv);
@@ -95,8 +96,10 @@ CAMLprim value netsys_alloc_memory_pages(value addrv, value pv)
     length = Int_val(pv);
     pgsize = sysconf(_SC_PAGESIZE);
     length = ((length - 1) / pgsize + 1) * pgsize;  /* fixup */
+    flags = PROT_READ|PROT_WRITE;
+    if (Bool_val(execv)) flags |= PROT_EXEC;
 
-    data = mmap(start, length, PROT_READ|PROT_WRITE, 
+    data = mmap(start, length, flags, 
 		MAP_PRIVATE | MAP_ANON, (-1), 0);
     if (data == (void *) -1) uerror("mmap", Nothing);
 
@@ -173,6 +176,7 @@ CAMLprim value netsys_map_file(value fdv,
     int fd, shared;
     off_t pos, savepos, eofpos, basize0;
           /* Att: pos might be 64 bit even on 32 bit systems! */
+    struct stat st;
     void *addr, *eff_addr;
     intnat size;
     uintnat basize;
@@ -180,6 +184,9 @@ CAMLprim value netsys_map_file(value fdv,
     char c;
     uintnat pagesize, delta;
 
+    /* Avoid here seeking at all costs. On some systems, shared memory
+       descriptors do neither allow seek, nor read/write.
+    */
     fd = Int_val(fdv);
     pos0 = Int64_val(posv);
     if (((int64) ((off_t) pos0)) != pos0)
@@ -192,10 +199,8 @@ CAMLprim value netsys_map_file(value fdv,
 
     pagesize = sysconf(_SC_PAGESIZE);
 
-    savepos = lseek(fd, 0, SEEK_CUR);
-    if (savepos == -1) uerror("lseek", Nothing);
-    eofpos = lseek(fd, 0, SEEK_END);
-    if (eofpos == -1) uerror("lseek", Nothing);
+    if (fstat(fd, &st) == -1) uerror("fstat", Nothing);
+    eofpos = st.st_size;
     
     if (size == -1) {
 	if (eofpos < pos) 
@@ -209,15 +214,11 @@ CAMLprim value netsys_map_file(value fdv,
 	if (size < 0)
 	    invalid_argument("netsys_map_file");
 	if (eofpos - pos < size) {
-	    if (lseek(fd, pos + size - 1, SEEK_SET) == -1)
-		uerror("lseek", Nothing);
-	    c = 0;
-	    if (write(fd, &c, 1) != 1) uerror("write", Nothing);
+	    if (ftruncate(fd, pos + size) == -1)
+		uerror("ftruncate", Nothing);
 	}
 	basize = size;
     }
-    lseek(fd, savepos, SEEK_SET);
-
     delta = (uintnat) (pos % pagesize);
     eff_addr = mmap(addr, basize + delta, PROT_READ | PROT_WRITE,
 		    shared, fd, pos - delta);
