@@ -1,4 +1,4 @@
-(* $Id: netsys_posix.ml 1742 2012-02-29 17:39:17Z gerd $ *)
+(* $Id: netsys_posix.ml 1790 2012-07-17 14:14:14Z gerd $ *)
 
 open Unix
 open Printf
@@ -702,9 +702,15 @@ let run_post_fork_handlers() =
 type at_flag = AT_EACCESS | AT_SYMLINK_NOFOLLOW | AT_REMOVEDIR
 
 (* The stubs assume these type definitions: *)
+IFDEF HAVE_O_SHARE_DELETE THEN
+type open_flag1 = Unix.open_flag =
+    O_RDONLY | O_WRONLY | O_RDWR | O_NONBLOCK | O_APPEND | O_CREAT | O_TRUNC
+  | O_EXCL | O_NOCTTY | O_DSYNC | O_SYNC | O_RSYNC | O_SHARE_DELETE
+ELSE
 type open_flag1 = Unix.open_flag =
     O_RDONLY | O_WRONLY | O_RDWR | O_NONBLOCK | O_APPEND | O_CREAT | O_TRUNC
   | O_EXCL | O_NOCTTY | O_DSYNC | O_SYNC | O_RSYNC
+ENDIF
 
 type access_permission1 = Unix.access_permission =
     R_OK | W_OK | X_OK | F_OK
@@ -835,23 +841,27 @@ let spawn ?(chdir = Wd_keep) ?(pg = Pg_keep) ?(fd_actions = [])
 	   (fun sa -> match sa with Sig_ignore _ -> true | _ -> false) 
 	   sig_actions) in
   
-  if use_posix_spawn then
+  try
+    if not use_posix_spawn then
+      failwith "USE_FORK_EXEC";
+
     netsys_posix_spawn pg fd_actions sig_actions env cmd args
-  else (
-    (* Fixup: if pg = Pg_new_fg_group, we remove any Sig_default for
-       SIGTTOU from sig_actions. Because of special handling, the effect
-       of Sig_default is enforced by the implementation, but this must be
-       done at [execve] time.
-     *)
-    let sig_actions =
-      if pg = Pg_new_fg_group then
-	List.filter 
-	  (fun spec -> spec <> Sig_default Sys.sigttou)
-	  sig_actions
-      else
-	sig_actions in
-    netsys_spawn chdir pg fd_actions sig_actions env cmd args
-  )
+      (* may also fail with "USE_FORK_EXEC" in some cases *)
+  with
+    | Failure "USE_FORK_EXEC" ->
+	(* Fixup: if pg = Pg_new_fg_group, we remove any Sig_default for
+	   SIGTTOU from sig_actions. Because of special handling, the effect
+	   of Sig_default is enforced by the implementation, but this must be
+	   done at [execve] time.
+	 *)
+	let sig_actions =
+	  if pg = Pg_new_fg_group then
+	    List.filter 
+	      (fun spec -> spec <> Sig_default Sys.sigttou)
+	      sig_actions
+	  else
+	    sig_actions in
+	netsys_spawn chdir pg fd_actions sig_actions env cmd args
 
 type watched_subprocess = 
     { atom_idx : int;
@@ -1220,7 +1230,8 @@ let shm_create prefix size =
 	with Unix.Unix_error(Unix.EINVAL,_,_) -> ()
 	  (* OSX seems to throw EINVAL here *)
       );
-      Unix.ftruncate fd size;
+      if size > 0 then
+        Unix.ftruncate fd size;
       (fd, name)
     with
       | Unix.Unix_error(Unix.EEXIST,_,_) ->
@@ -1239,7 +1250,7 @@ type 'sem_kind semaphore =
        *)
 
 type named_semaphore = [ `Named ] semaphore
-type anon_sempahore = [ `Anonymous ] semaphore
+type anon_semaphore = [ `Anonymous ] semaphore
 
 type sem_open_flag =
   | SEM_O_CREAT
@@ -1252,8 +1263,14 @@ type sem_wait_behavior =
 let dummy_mem =
   Bigarray.Array1.create Bigarray.char Bigarray.c_layout 0
 
-external have_posix_semaphores : unit -> bool 
-  = "netsys_have_sem"
+external have_anon_posix_semaphores : unit -> bool 
+  = "netsys_have_sem_anon"
+
+external have_named_posix_semaphores : unit -> bool 
+  = "netsys_have_sem_named"
+
+let have_posix_semaphores() =
+  have_anon_posix_semaphores() && have_named_posix_semaphores()
 
 external netsys_sem_size : unit -> int
   = "netsys_sem_size"
